@@ -62,12 +62,12 @@ void KTextToHTMLTest::testGetEmailAddress()
     KTextToHTMLHelper ll6(QStringLiteral(" @bar.baz"), 1);
     QVERIFY(ll6.getEmailAddress().isEmpty());
     KTextToHTMLHelper ll7(QStringLiteral(".!#$%&'*+-/=?^_`{|}~@bar.baz"),
-                    qstrlen(".!#$%&'*+-/=?^_`{|}~"));
+                          qstrlen(".!#$%&'*+-/=?^_`{|}~"));
     QVERIFY(ll7.getEmailAddress().isEmpty());
 
     // allowed special chars in local part of address
     KTextToHTMLHelper ll8(QStringLiteral("a.!#$%&'*+-/=?^_`{|}~@bar.baz"),
-                    qstrlen("a.!#$%&'*+-/=?^_`{|}~"));
+                          qstrlen("a.!#$%&'*+-/=?^_`{|}~"));
     QCOMPARE(ll8.getEmailAddress(), QStringLiteral("a.!#$%&'*+-/=?^_`{|}~@bar.baz"));
 
     // '@' in domain part
@@ -157,7 +157,7 @@ void KTextToHTMLTest::testGetUrl2(const QString &left, const QString &right)
     urls << QStringLiteral("user:pass@www.kde.org:1234/sub/path:with:colon/special(123)?a=1#anchor[bla]");
     urls << QStringLiteral("user:pass@www.kde.org:1234/\nsub/path:with:colon/\nspecial(123)?\na=1#anchor[bla]");
     urls << QStringLiteral("user:pass@www.kde.org:1234/  \n  sub/path:with:colon/  \n\t   \t   special(123)?") +
-                          QStringLiteral("\n\t  \n\t   a=1#anchor[bla]");
+            QStringLiteral("\n\t  \n\t   a=1#anchor[bla]");
 
     foreach (const QString &schema, schemas) {
         foreach (QString url, urls) {   //krazy:exclude=foreach
@@ -202,7 +202,7 @@ void KTextToHTMLTest::testGetUrl2(const QString &left, const QString &right)
     urlsWithoutSchema << QStringLiteral(".kde.org:1234/sub/path:with:colon/special(123)?a=1#anchor[bla]");
     urlsWithoutSchema << QStringLiteral(".kde.org:1234/\nsub/path:with:colon/\nspecial(123)?\na=1#anchor[bla]");
     urlsWithoutSchema << QStringLiteral(".kde.org:1234/  \n  sub/path:with:colon/  \n\t   \t   special(123)?") +
-                                       QStringLiteral("\n\t  \n\t   a=1#anchor[bla]");
+                         QStringLiteral("\n\t  \n\t   a=1#anchor[bla]");
 
     QStringList starts;
     starts << QStringLiteral("www") << QStringLiteral("ftp") << QStringLiteral("news:www");
@@ -240,7 +240,7 @@ void KTextToHTMLTest::testGetUrl2(const QString &left, const QString &right)
     }
 
     // test max url length
-    QString url = QStringLiteral("http://www.kde.org/this/is/a_very_loooooong_url/test/test/test");
+    QString url = QStringLiteral("https://www.kde.org/this/is/a_very_loooooong_url/test/test/test");
     {
         KTextToHTMLHelper ll(url, 0, 10);
         QVERIFY(ll.getUrl().isEmpty());    // url too long
@@ -372,62 +372,126 @@ void KTextToHTMLTest::testHtmlConvert_data()
                                  "www.example.com/test.cpp</a>.";
 
     // Bug 313719 - URL in parenthesis
-    QTest::newRow("url-in-parenthesis-1") << "KDE (website http://www.kde.org)"
+    QTest::newRow("url-in-parenthesis-1") << "KDE (website https://www.kde.org)"
                                           << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                                          << "KDE (website <a href=\"http://www.kde.org\">http://www.kde.org</a>)";
-    QTest::newRow("url-in-parenthesis-2") << "KDE website (http://www.kde.org)"
+                                          << "KDE (website <a href=\"https://www.kde.org\">https://www.kde.org</a>)";
+    QTest::newRow("url-in-parenthesis-2") << "KDE website (https://www.kde.org)"
                                           << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                                          << "KDE website (<a href=\"http://www.kde.org\">http://www.kde.org</a>)";
-    QTest::newRow("url-in-parenthesis-3") << "bla (http://www.kde.org - section 5.2)"
+                                          << "KDE website (<a href=\"https://www.kde.org\">https://www.kde.org</a>)";
+    QTest::newRow("url-in-parenthesis-3") << "bla (https://www.kde.org - section 5.2)"
                                           << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                                          << "bla (<a href=\"http://www.kde.org\">http://www.kde.org</a> - section 5.2)";
+                                          << "bla (<a href=\"https://www.kde.org\">https://www.kde.org</a> - section 5.2)";
     
-   // Fix url as foo <<url> <url>> when we concatened them.
-   QTest::newRow("url-with-url") << "foo <http://www.kde.org/ <http://www.kde.org/>>"
-                               << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                               << "foo &lt;<a href=\"http://www.kde.org/ \">http://www.kde.org/ </a>&lt;<a href=\"http://www.kde.org/\">http://www.kde.org/</a>&gt;&gt;";
+    // Fix url as foo <<url> <url>> when we concatened them.
+    QTest::newRow("url-with-url") << "foo <https://www.kde.org/ <https://www.kde.org/>>"
+                                  << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                  << "foo &lt;<a href=\"https://www.kde.org/ \">https://www.kde.org/ </a>&lt;<a href=\"https://www.kde.org/\">https://www.kde.org/</a>&gt;&gt;";
 
-   //Fix url exploit
-   QTest::newRow("url-exec-html") << "https://\"><!--"
-                               << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                               << "https://&quot;&gt;&lt;!--";
+    //Fix url exploit
+    QTest::newRow("url-exec-html") << "https://\"><!--"
+                                   << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                   << "https://&quot;&gt;&lt;!--";
 
-   QTest::newRow("url-exec-html-2") << "https://192.168.1.1:\"><!--"
-                               << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                               << "https://192.168.1.1:&quot;&gt;&lt;!--";
+    QTest::newRow("url-exec-html-2") << "https://192.168.1.1:\"><!--"
+                                     << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                     << "https://192.168.1.1:&quot;&gt;&lt;!--";
 
-   QTest::newRow("url-exec-html-3") << "https://<IP>:\"><!--"
-                               << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                               << "https://&lt;IP&gt;:&quot;&gt;&lt;!--";
+    QTest::newRow("url-exec-html-3") << "https://<IP>:\"><!--"
+                                     << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                     << "https://&lt;IP&gt;:&quot;&gt;&lt;!--";
 
-   QTest::newRow("url-exec-html-4") << "https://<IP>:/\"><!--"
-                               << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                               << "https://&lt;IP&gt;:/&quot;&gt;&lt;!--";
+    QTest::newRow("url-exec-html-4") << "https://<IP>:/\"><!--"
+                                     << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                     << "https://&lt;IP&gt;:/&quot;&gt;&lt;!--";
 
-   QTest::newRow("url-exec-html-5") << "https://<IP>:/\"><script>alert(1);</script><!--"
-                               << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                               << "https://&lt;IP&gt;:/&quot;&gt;&lt;script&gt;alert(1);&lt;/script&gt;&lt;!--";
+    QTest::newRow("url-exec-html-5") << "https://<IP>:/\"><script>alert(1);</script><!--"
+                                     << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                     << "https://&lt;IP&gt;:/&quot;&gt;&lt;script&gt;alert(1);&lt;/script&gt;&lt;!--";
 
-   QTest::newRow("url-exec-html-6") << "https://<IP>:/\"><script>alert(1);</script><!--\nTest2"
-                               << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                               << "https://&lt;IP&gt;:/&quot;&gt;&lt;script&gt;alert(1);&lt;/script&gt;&lt;!--\nTest2";
+    QTest::newRow("url-exec-html-6") << "https://<IP>:/\"><script>alert(1);</script><!--\nTest2"
+                                     << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                     << "https://&lt;IP&gt;:/&quot;&gt;&lt;script&gt;alert(1);&lt;/script&gt;&lt;!--\nTest2";
 
 
-   QTest::newRow("url-with-ref-in-[") << "https://www.kde.org[1]"
-                               << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                               << "<a href=\"https://www.kde.org\">https://www.kde.org</a>[1]";
+    QTest::newRow("url-with-ref-in-[") << "https://www.kde.org[1]"
+                                       << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                       << "<a href=\"https://www.kde.org\">https://www.kde.org</a>[1]";
 
-   QTest::newRow("url-with-ref-in-[2") << "[http://www.example.org/][whatever]"
-                               << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                               << "[<a href=\"http://www.example.org/\">http://www.example.org/</a>][whatever]";
-   //Bug 346132
-   QTest::newRow("url-with-ref-in-<") << "http://www.foo.bar<http://foo.bar/>"
-                               << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                               << "<a href=\"http://www.foo.bar\">http://www.foo.bar</a>[1]";
+    QTest::newRow("url-with-ref-in-[2") << "[http://www.example.org/][whatever]"
+                                        << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                        << "[<a href=\"http://www.example.org/\">http://www.example.org/</a>][whatever]";
+    //Bug 346132
+    QTest::newRow("url-with-ref-in-<") << "http://www.foo.bar<http://foo.bar/>"
+                                       << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                       << "<a href=\"http://www.foo.bar\">http://www.foo.bar</a>&lt;<a href=\"http://foo.bar/\">http://foo.bar/</a>&gt;";
 
-   QTest::newRow("url-with-ref-in-]") << "[Please visit our booth 24-25 http://example.com/]"
-                               << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
-                               << "[Please visit our booth 24-25 <a href=\"http://example.com/\">http://example.com/</a>]";
+    QTest::newRow("url-with-ref-in-]") << "[Please visit our booth 24-25 http://example.com/]"
+                                       << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                       << "[Please visit our booth 24-25 <a href=\"http://example.com/\">http://example.com/</a>]";
+
+    QTest::newRow("two url with space") << "http://www.kde.org/standards/kcfg/1.0 http://www.kde.org/"
+                                       << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                       << "<a href=\"http://www.kde.org/standards/kcfg/1.0\">http://www.kde.org/standards/kcfg/1.0</a> <a href=\"http://www.kde.org/\">http://www.kde.org/</a>";
+
+
+    //Bug kmail
+    QTest::newRow("two url with space-2") << "@@ -55,6 +55,10 @@ xsi:schemaLocation=\"http://www.kde.org/standards/kcfg/1.0 http://www.kde.org/"
+                                       << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+                                       << "@@ -55,6 +55,10 @@ xsi:schemaLocation=&quot;<a href=\"http://www.kde.org/standards/kcfg/1.0\">http://www.kde.org/standards/kcfg/1.0</a> <a href=\"http://www.kde.org/\">http://www.kde.org/</a>";
+
+    const auto opt = KTextToHTML::PreserveSpaces | KTextToHTML::ConvertPhoneNumbers;
+    // tel: urls
+    QTest::newRow("tel url compact") << "bla bla <tel:+491234567890> bla bla" << opt
+        << "bla bla &lt;<a href=\"tel:+491234567890\">tel:+491234567890</a>&gt; bla bla";
+    QTest::newRow("tel url fancy") << "bla bla tel:+49-321-123456 bla bla" << opt
+        << "bla bla <a href=\"tel:+49-321-123456\">tel:+49-321-123456</a> bla bla";
+
+    // negative tel: url tests
+    QTest::newRow("empty tel url") << "bla tel: blub" << opt
+        << "bla tel: blub";
+
+    // phone numbers
+    QTest::newRow("tel compact international") << "call +49123456789, then hang up" << opt
+        << "call <a href=\"tel:+49123456789\">+49123456789</a>, then hang up";
+    QTest::newRow("tel parenthesis/spaces international") << "phone:+33 (01) 12 34 56 78 blub" << opt
+        << "phone:<a href=\"tel:+330112345678\">+33 (01) 12 34 56 78</a> blub";
+    QTest::newRow("tel dashes international") << "bla +44-321-1-234-567" << opt
+        << "bla <a href=\"tel:+443211234567\">+44-321-1-234-567</a>";
+    QTest::newRow("tel dashes/spaces international") << "+1 123-456-7000 blub" << opt
+        << "<a href=\"tel:+11234567000\">+1 123-456-7000</a> blub";
+    QTest::newRow("tel spaces international") << "bla +32 1 234 5678 blub" << opt
+        << "bla <a href=\"tel:+3212345678\">+32 1 234 5678</a> blub";
+    QTest::newRow("tel slash domestic") << "bla 030/12345678 blub" << opt
+        << "bla <a href=\"tel:03012345678\">030/12345678</a> blub";
+    QTest::newRow("tel slash/space domestic") << "Tel.: 089 / 12 34 56 78" << opt
+        << "Tel.: <a href=\"tel:08912345678\">089 / 12 34 56 78</a>";
+    QTest::newRow("tel follow by parenthesis") << "Telefon: 0 18 05 / 12 23 46 (14 Cent/Min.*)" << opt
+        << "Telefon: <a href=\"tel:01805122346\">0 18 05 / 12 23 46</a> (14 Cent/Min.*)";
+    QTest::newRow("tel space single digit at end") << "0123/123 456 7" << opt
+        << "<a href=\"tel:01231234567\">0123/123 456 7</a>";
+    QTest::newRow("tel space around dash") << "bla +49 (0) 12 23 - 45 6000 blub" << opt
+        << "bla <a href=\"tel:+4901223456000\">+49 (0) 12 23 - 45 6000</a> blub";
+    QTest::newRow("tel two numbers speparated by dash") << "bla +49 (0) 12 23 46 78 - +49 0123/123 456 78 blub" << opt
+        << "bla <a href=\"tel:+49012234678\">+49 (0) 12 23 46 78</a> - <a href=\"tel:+49012312345678\">+49 0123/123 456 78</a> blub";
+
+    // negative tests for phone numbers
+    QTest::newRow("non-tel number") << "please send 1200 cakes" << opt
+        << "please send 1200 cakes";
+    QTest::newRow("non-tel alpha-numeric") << "bla 1-123-456-ABCD blub" << opt
+        << "bla 1-123-456-ABCD blub";
+    QTest::newRow("non-tel alpha prefix") << "ABCD0123-456-789" << opt
+        << "ABCD0123-456-789";
+    QTest::newRow("non-tel date") << "bla 02/03/2019 blub" << opt
+        << "bla 02/03/2019 blub";
+    QTest::newRow("non-tel too long") << "bla +012-4567890123456 blub" << opt
+        << "bla +012-4567890123456 blub";
+    QTest::newRow("non-tel unbalanced") << "bla +012-456789(01 blub" << opt
+        << "bla +012-456789(01 blub";
+    QTest::newRow("non-tel nested") << "bla +012-4(56(78)90)1 blub" << opt
+        << "bla +012-4(56(78)90)1 blub";
+    QTest::newRow("tel extraction disabled") << "call +49123456789 now"
+        << KTextToHTML::Options(KTextToHTML::PreserveSpaces)
+        << "call +49123456789 now";
 }
 
 
@@ -440,9 +504,6 @@ void KTextToHTMLTest::testHtmlConvert()
     QEXPECT_FAIL("punctation-bug", "Linklocator does not properly detect punctation as boundaries",
                  Continue);
 
-    QEXPECT_FAIL("url-with-ref-in-<", "Linklocator does not properly parse url with <",
-                 Continue);
-
     QString actualHtml = KTextToHTML::convertToHtml(plainText, flags);
     QCOMPARE(actualHtml, htmlText);
 }
@@ -452,7 +513,7 @@ void KTextToHTMLTest::benchHtmlConvert_data()
     QTest::addColumn<QString>("text");
     QTest::addColumn<KTextToHTML::Options>("options");
 
-    auto text = QString("foo bar asdf :)").repeated(1000);
+    auto text = QString(QStringLiteral("foo bar asdf :)")).repeated(1000);
     QTest::newRow("plain") << text << KTextToHTML::Options();
     QTest::newRow("preserve-spaces") << text << KTextToHTML::Options(KTextToHTML::PreserveSpaces);
     QTest::newRow("highlight-text") << text << KTextToHTML::Options(KTextToHTML::HighlightText);
